/*
 * Created By Drake Vidkjer
 * Created on 10-19-17
 * Modified on 10-30-17
 * Software implementation of floating point arithmetic
 */
 
#define F64_EXP_MASK 0x7FF
#define F64_EXP_SHIFT 52
#define F64_MANT_MASK 0xFFFFFFFFFFFFF
#define F64_MANT_SHIFT 0

double normalize(double arg, int* exp)
{
	//get a pointer to use with bitshifting since cant bit shift on floating point #
	//unsigned long* ptr = (unsigned long*)&arg;
	/*
	typedef union
	{
		double fv;
		unsigned long iv;
	} longDoub;
	longDoub temp;
	temp.fv = arg;
	* */
	unsigned long* iptr = (unsigned long*)&arg;
	
	int sign;
	long te;
	double a;
	
	//if +/-infinity then return +/- infinity
	if(arg == INFINITY)
	{
		return INFINITY;
	}
	else if(arg == -INFINITY)
	{
		return -INFINITY;
	}
	//if NaN then return NaN
	else if(arg == NAN)
	{
		return NAN;
	}
	//if zero then return zero
	else if(arg == 0)
	{
		return 0;
		*exp = 0;
	}
	//else do what we need to do
	else
	{
		//get the sign bit
		sign = (*iptr >> 63) & 1;
		//printf("sign = %d", sign);
		//get the exponent and then normalize it
		*exp = (*iptr >> F64_EXP_SHIFT) & F64_EXP_MASK;
		*exp -= 1023;
		
		//need to add one to the exponent value because of how we shift the binary point
		//when getting the mantissa
		*exp += 1;
		
		//mask off everything but the mantissa
		te = (*iptr & F64_MANT_MASK);
		
		//maniplate the mantissa to make it work how we want it.
		//To do this, I think i would have to take the mantissa, add a 1 ti the beginning
		//and move the binary point to right after that one. I don't exactly know how to do
		//that but I think it would involve anding with a leading 1 and then multiply by 
		//2^-52
		te = te | 0b10000000000000000000000000000000000000000000000000000;
		a = te * pow(2, -52);
		a = a /2;
		//set the sign correctly
		return ((1 - (2 * sign)) * a);
	}
	return -1;
}

unsigned long getExp (double val)  
{
         unsigned long* iptr = (unsigned long*)&val;
         
         return (((*iptr >> F64_EXP_SHIFT) & F64_EXP_MASK)-1023);
}
 
unsigned long getMan(double val) 
{
         unsigned long* iptr = (unsigned long*)&val;
         if( *iptr == 0 )
         {
			 return 0;
		 }
		 
         return ((*iptr & F64_MANT_MASK) | 0x8000000000000000);        
}

double add(double arg1, double arg2)
{
	//define variables
	unsigned long litMan, bigMan, res;
	unsigned long litExp, bigExp, exp1, exp2;
	long expDif, manSum, manShift;
	double fres;
	
	//edge cases
	if(arg1 == INFINITY || arg2 == INFINITY)
	{
		return INFINITY;
	}
	else if(arg1 == -INFINITY || arg2 == -INFINITY)
	{
		return -INFINITY;
	}
	else if(arg1 == NAN || arg2 == NAN)
	{
		return NAN;
	}
	
	//get the exponents
	exp1 = getExp(arg1);
	exp2 = getExp(arg2);
	
	//decide which value is bigger and set variables as such
	if(exp1 > exp2)
	{
		litExp = exp2;
		litMan = getMan(arg2);
		bigExp = exp1;
		bigMan = getMan(arg1);
	}
	else
	{
		litExp = exp1;
		litMan = getMan(arg1);
		bigExp = exp2;
		bigMan = getMan(arg2);
	}
	
	//get the exponent difference and shift little manitssa with it
	expDif = bigExp - litExp;
	manShift = litMan >> expDif;
	
	//add the mantissas
	manSum = bigMan + manShift;
	
	//convert back to format we can use
	res = (((bigExp + 1023) << F64_EXP_SHIFT) | (manSum & F64_MANT_MASK));
	fres = *(double*)&res;
	return fres;
}
